//
//  Planet.swift
//  Solar system in Metal
//
//  Created by Albert Kwaśkiewicz on 5/4/20.
//  Copyright © 2020 AC/DC. All rights reserved.
//

import MetalKit

enum PlanetName {
    case mercury, venus, earth, mars, jupiter, saturn, uranus, neptune
}

// all distances and lengths are presented as kilometers
class Planet: Node {
    let device = Renderer.device!
    let pipelineState: MTLRenderPipelineState
    var meshes: [MTKMesh] = []
    var mdlMesh: MDLMesh!
    var mesh: MTKMesh!
    
    let scaleFactor: Float = 10_000
    
    var diameter: Float
    private var radius: Float {
        return diameter / 2.0
    }
    var distance: Float
    var moons: Int
    
    var size: Float {
        return radius / scaleFactor
    }
        
    init(planet: PlanetName) {
        
        mdlMesh = Planet.createPlanet(device: device, size: 1)
        do {
            mesh = try MTKMesh(mesh: mdlMesh, device: device)
        } catch let error {
            print(error.localizedDescription)
        }
      
        meshes.append(mesh)
        
        switch planet {
        case .mercury:
            diameter = 4_879.4
            self.distance = 57_909_00.0
            self.moons = 0
        case .venus:
            diameter = 12_103.6
            self.distance = 108_160_000.0
            self.moons = 0
        case .earth:
            diameter = 12_756.3
            self.distance = 149_600_000.0
            self.moons = 1
        case .mars:
            self.diameter =  6_792.4
            self.distance = 227_990_00.0
            self.moons = 2
        case .jupiter:
            self.diameter = 142_984.0
            self.distance = 778_360_000.0
            self.moons = 79
        case .saturn:
            self.diameter = 120_536.0
            self.distance = 1_433_500_500.0
            self.moons = 62
        case .uranus:
            self.diameter = 51_118.0
            self.distance = 2_872_400_000.0
            self.moons = 27
        case .neptune:
            self.diameter = 49_528.0
            self.distance = 4_498_400_000.0
            self.moons = 14
        }
        
        pipelineState = Planet.buildPipelineState()
        
        super.init()
        
//        guard let device = Renderer.device else {
//            return
//        }
        
        switch planet {
        case .mercury:
            self.name = "Mercury"
        case .venus:
            self.name = "Venus"
        case .earth:
            self.name = "Earth"
        case .mars:
            name = "Mars"
        case .jupiter:
            name = "Jupiter"
        case .saturn:
            name = "Saturn"
        case .uranus:
            name = "Uranus"
        case .neptune:
            name = "Neptune"
        }
      mesh.name = name 
    }
}

extension Planet {
    static func createPlanet(device: MTLDevice, size: Float) -> MDLMesh {
        let allocator = MTKMeshBufferAllocator(device: device)
        let mesh = MDLMesh(sphereWithExtent: [size, size, size],
                           segments: [10, 10],
                           inwardNormals: false,
                           geometryType: .triangles,
                           allocator: allocator)
        return mesh
    }
    
    static func buildPipelineState() -> MTLRenderPipelineState {
        let library = Renderer.library
        let vertexFunction = library?.makeFunction(name: "vertex_main")
        let fragmentFunction = library?.makeFunction(name: "fragment_main")
        
        var pipelineState: MTLRenderPipelineState
        let pipelineDescriptor = MTLRenderPipelineDescriptor()
        pipelineDescriptor.vertexFunction = vertexFunction
        pipelineDescriptor.fragmentFunction = fragmentFunction
        let vertexDescriptor = MDLVertexDescriptor.defaultVertexDescriptor
        pipelineDescriptor.vertexDescriptor = MTKMetalVertexDescriptorFromModelIO(vertexDescriptor)
        pipelineDescriptor.colorAttachments[0].pixelFormat = .bgra8Unorm
        pipelineDescriptor.depthAttachmentPixelFormat = .depth32Float
        do {
            pipelineState = try Renderer.device.makeRenderPipelineState(descriptor: pipelineDescriptor)
        } catch let error {
            fatalError(error.localizedDescription)
        }
        return pipelineState
    }
}

class Sun: Node {
    
    let childPlanets: [Planet] = []
    let scaleFactor: Float = 10_000
    let diameter: Float = 1_392_684.0
    
    var radius: Float {
        return diameter / 2.0
    }
    
    var size: Float
    {
        return radius / scaleFactor
    }
    
    func createSun(device: MTLDevice) -> MDLMesh {
        let allocator = MTKMeshBufferAllocator(device: device)
        let mesh = MDLMesh(sphereWithExtent: [size, size, size],
                           segments: [1000, 1000],
                           inwardNormals: false,
                           geometryType: .points,
                           allocator: allocator)
        return mesh
    }
}

class Primitive {
    static func makeCube(device: MTLDevice, size: Float) -> MDLMesh {
        let allocator = MTKMeshBufferAllocator(device: device)
        let mesh = MDLMesh(boxWithExtent: [size, size, size],
                           segments: [1, 1, 1],
                           inwardNormals: false,
                           geometryType: .triangles,
                           allocator: allocator)
        
        
        return mesh
    }
    
    static func makeSphere(device: MTLDevice, size: Float) -> MDLMesh {
        let allocator = MTKMeshBufferAllocator(device: device)
        let mesh = MDLMesh(sphereWithExtent: [size, size, size],
                           segments: [1000, 1000],
                           inwardNormals: false,
                           geometryType: .lines,
                           allocator: allocator)
        return mesh
    }
}
